local AddonName = "TomeOfTeleportation"
local AddonTitle = "Tome of Teleportation"
-- Special case strings start with number to force them to be sorted first.
local HearthString = "0 Hearth"
local RecallString = "1 Astral Recall"

local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local icon = LibStub("LibDBIcon-1.0")
local dataobj = ldb:NewDataObject("TomeTele", {
	label = "TomeOfTeleportation", 
	type = "data source", 
	icon = "Interface\\Icons\\Spell_Arcane_TeleportDalaran", 
	text = "Teleport"
})

local TeleporterParentFrame = nil
local CastSpell = nil
local ItemSlot = nil
local OldItems = {}
local ButtonSettings = {}
local IsVisible = false
local ShouldNotBeEquiped = {}
local ShouldBeEquiped = {}
local EquipTime = 0

_G["BINDING_HEADER_TOMEOFTELEPORTATION"] = "Tome of Teleportation"

local InvTypeToSlot = 
{	
	["INVTYPE_HEAD"] = 1,
	["INVTYPE_NECK"] = 2,
	["INVTYPE_SHOULDER"] = 3,
	["INVTYPE_BODY"] = 4,
	["INVTYPE_CHEST"] = 5,
	["INVTYPE_ROBE"] = 5,
	["INVTYPE_WAIST"] = 6,
	["INVTYPE_LEGS"] = 7,
	["INVTYPE_FEET"] = 8,
	["INVTYPE_WRIST"] = 9,
	["INVTYPE_HAND"] = 10,
	["INVTYPE_FINGER"] = 11,
	["INVTYPE_TRINKET"] = 13,
	["INVTYPE_CLOAK"] = 15,
	["INVTYPE_2HWEAPON"] = 16,
	["INVTYPE_WEAPONMAINHAND"] = 16,
	["INVTYPE_TABARD"] = 19
}

local MapIDAlteracValley = 401
local MapIDIsleOfThunder = 928
local MapIDDalaran = 504
local MapIDTanaanJungle = 945

local ContinentIdPandaria = 6
local ContinentIdDraenor = 7

local ST_Item = 1
local ST_Spell = 2
local ST_Challenge = 3

local SpellNameIndex = 6

local DefaultOptions = 
{
	["scale"] = { 1, "Scale to apply to the whole window" },
	["buttonHeight"] = { 26, "Height of each button" },
	["buttonWidth"] = { 128, "Width of each button" },
	["labelHeight"] = { 16, "Height of the button labels" },
	["maximumHeight"] = { 200, "Maximum height of the main window" },
	["fontHeight"] = { 10, "The height of the button font" },
	["buttonInset"] = { 6, nil },
	["showHelp"] = { false, "Display the help text (true or false)" }
}

-- { id, isItem, destination, zoneLimited, consumable, spellName }
-- It probably won't work if a single player has two different items
-- with the same name in their inventory, but I don't think that's possible.
-- spellName will be filled in when the addon loads.
local TeleporterSpells = 
{	
	{ 93672, ST_Item, HearthString },		-- Dark Portal
	{ 54452, ST_Item, HearthString },		-- Ethereal Portal
	{ 6948, ST_Item, HearthString },		-- Hearthstone
	{ 28585, ST_Item, HearthString },		-- Ruby Slippers
	{ 37118, ST_Item, HearthString, nil, true },	-- Scroll of Recall
	{ 44314, ST_Item, HearthString, nil, true },	-- Scroll of Recall II
	{ 44315, ST_Item, HearthString, nil, true },	-- Scroll of Recall III
	{ 64488, ST_Item, HearthString },		-- The Innkeeper's Daughter	
	
	{ 556, ST_Spell, RecallString },		-- Astral Recall
	
	-- Alterac Valley
	{ 17690, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Frostwolf Insignia Rank 1
	{ 17905, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Frostwolf Insignia Rank 2
	{ 17906, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Frostwolf Insignia Rank 3
	{ 17907, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Frostwolf Insignia Rank 4
	{ 17908, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Frostwolf Insignia Rank 5
	{ 17909, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Frostwolf Insignia Rank 6
	{ 17691, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Stormpike Insignia Rank 1
	{ 17900, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Stormpike Insignia Rank 2
	{ 17901, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Stormpike Insignia Rank 3
	{ 17902, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Stormpike Insignia Rank 4
	{ 17903, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Stormpike Insignia Rank 5
	{ 17904, ST_Item, "Alterac Valley", MapIDAlteracValley },	-- Stormpike Insignia Rank 6
	
	{ 116413, ST_Item, "Ashran", nil, true },	-- Scroll of Town Portal
	{ 119183, ST_Item, "Ashran", nil, true },	-- Scroll of Risky Recall
	{ 176246, ST_Spell, "Ashran" },			-- Portal: Stormshield
	{ 176248, ST_Spell, "Ashran" },			-- Teleport: Stormshield
	{ 176244, ST_Spell, "Ashran" },			-- Portal: Warspear
	{ 176242, ST_Spell, "Ashran" },			-- Teleport: Warspear

	{ 95051, ST_Item, "Bizmo's Brawlpub" },	-- The Brassiest Knuckle
	{ 118907, ST_Item, "Bizmo's Brawlpub" },	-- Pit Fighter's Punching Ring

	{ 32757, ST_Item, "Black Temple" },	-- Blessed Medallion of Karabor
	
	{ 37863, ST_Item, "Blackrock Depths" },-- Direbrew's Remote
	
	{ 169771, ST_Challenge, "Blackrock Foundry" },		-- Teleport: Blackrock Foundry
	
	{ 30544, ST_Item, "Blade's Edge" },	-- Ultrasafe Transporter - Toshley's Station
	
	{ 118662, ST_Item, "Bladespire Fortress" }, -- Bladespire Relic
	
	{ 50287, ST_Item, "Booty Bay" },		-- Boots of the Bay
	
	{ 95050, ST_Item, "Brawl'gar Arena" },	-- The Brassiest Knuckle
	{ 118908, ST_Item, "Brawl'gar Arena" },-- Pit Fighter's Punching Ring

	{ 53140, ST_Spell, "Dalaran" },		-- Teleport: Dalaran
	{ 53142, ST_Spell, "Dalaran" },		-- Portal: Dalaran
	-- ilvl 200 rings
	{ 40586, ST_Item, "Dalaran" },			-- Band of the Kirin Tor
	{ 44934, ST_Item, "Dalaran" },			-- Loop of the Kirin Tor
	{ 44935, ST_Item, "Dalaran" },			-- Ring of the Kirin Tor
	{ 40585, ST_Item, "Dalaran" },			-- Signet of the Kirin Tor
	-- ilvl 213 rings
	{ 45688, ST_Item, "Dalaran" },			-- Inscribed Band of the Kirin Tor
	{ 45689, ST_Item, "Dalaran" },			-- Inscribed Loop of the Kirin Tor
	{ 45690, ST_Item, "Dalaran" },			-- Inscribed Ring of the Kirin Tor
	{ 45691, ST_Item, "Dalaran" },			-- Inscribed Signet of the Kirin Tor
	-- ilvl 226 rings
	{ 48954, ST_Item, "Dalaran" },			-- Etched Band of the Kirin Tor
	{ 48955, ST_Item, "Dalaran" },			-- Etched Loop of the Kirin Tor
	{ 48956, ST_Item, "Dalaran" },			-- Etched Ring of the Kirin Tor
	{ 48957, ST_Item, "Dalaran" },			-- Etched Signet of the Kirin Tor
	-- ilvl 251 rings
	{ 51560, ST_Item, "Dalaran" },			-- Runed Band of the Kirin Tor
	{ 51558, ST_Item, "Dalaran" },			-- Runed Loop of the Kirin Tor
	{ 51559, ST_Item, "Dalaran" },			-- Runed Ring of the Kirin Tor
	{ 51557, ST_Item, "Dalaran" },			-- Runed Signet of the Kirin Tor
	
	{ 43824, ST_Item, "Dalaran", MapIDDalaran },	-- The Schools of Arcane Magic - Mastery
	{ 52251, ST_Item, "Dalaran" },			-- Jaina's Locket
	
	{ 120145, ST_Spell, "Dalaran Crater" },-- Ancient Teleport: Dalaran
	{ 120146, ST_Spell, "Dalaran Crater" },-- Ancient Portal: Dalaran

	{ 3565, ST_Spell, "Darnassus" },		-- Teleport: Darnassus
	{ 11419, ST_Spell, "Darnassus" },		-- Portal: Darnassus
	
	{ 58487, ST_Item, "Deepholm", nil, true },	-- Potion of Deepholm
	
	{ 117389, ST_Item, "Draenor", -ContinentIdDraenor, true }, -- Draenor Archaeologist's Lodestone
	{ 112059, ST_Item, "Draenor" },				-- Wormhole Centrifuge
	{ 129929, ST_Item, "Draenor" },				-- Ever-Shifting Mirror
	
	{ 159897, ST_Challenge, "Draenor Dungeons" },	-- Teleport: Auchindoun
	{ 159895, ST_Challenge, "Draenor Dungeons" },	-- Teleport: Bloodmaul Slag Mines
	{ 159901, ST_Challenge, "Draenor Dungeons" },	-- Teleport: Overgrown Outpost
	{ 159900, ST_Challenge, "Draenor Dungeons" },	-- Teleport: Grimrail Depot
	{ 159896, ST_Challenge, "Draenor Dungeons" },	-- Teleport: Iron Docks
	{ 159899, ST_Challenge, "Draenor Dungeons" },	-- Teleport: Shadowmoon Burial Grounds
	{ 159898, ST_Challenge, "Draenor Dungeons" },	-- Teleport: Skyreach
	{ 159902, ST_Challenge, "Draenor Dungeons" },	-- Teleport: Upper Blackrock Spire

	{ 50977, ST_Spell, "Ebon Hold" },		-- Death Gate
	
	{ 110560, ST_Item, "Garrison" },		-- Garrison Hearthstone
	
	{ 32271, ST_Spell, "Exodar" },			-- Teleport: Exodar
	{ 32266, ST_Spell, "Exodar" },			-- Portal: Exodar
	
	{ 46874, ST_Item, "Icecrown" },		-- Argent Crusader's Tabard
	
	{ 3562, ST_Spell, "Ironforge" },		-- Teleport: Ironforge
	{ 11416, ST_Spell, "Ironforge" },		-- Portal: Ironforge
	
	{ 95567, ST_Item, "Isle of Thunder", MapIDIsleOfThunder },	-- Kirin Tor Beacon
	{ 95568, ST_Item, "Isle of Thunder", MapIDIsleOfThunder },	-- Sunreaver Beacon

	{ 118663, ST_Item, "Karabor" },		-- Relic of Karabor
	
	{ 22589, ST_Item, "Karazhan" },		-- Atiesh, Greatstaff of the Guardian
	{ 22630, ST_Item, "Karazhan" },		-- Atiesh, Greatstaff of the Guardian
	{ 22631, ST_Item, "Karazhan" },		-- Atiesh, Greatstaff of the Guardian
	{ 22632, ST_Item, "Karazhan" },		-- Atiesh, Greatstaff of the Guardian
	
	{ 18960, ST_Spell, "Moonglade" },		-- Teleport: Moonglade

	{ 30542, ST_Item, "Netherstorm" },		-- Dimensional Ripper - Area 52

	{ 48933, ST_Item, "Northrend" },		-- Wormhole Generator: Northrend

	{ 3567, ST_Spell, "Orgrimmar" },		-- Teleport: Orgrimmar
	{ 11417, ST_Spell, "Orgrimmar" },		-- Portal: Orgrimmar
	{ 63207, ST_Item, "Orgrimmar" },		-- Wrap of Unity
	{ 63353, ST_Item, "Orgrimmar" },		-- Shroud of Cooperation
	{ 65274, ST_Item, "Orgrimmar" },		-- Cloak of Coordination
	
	{ 87548, ST_Item, "Pandaria", -ContinentIdPandaria, true }, -- Lorewalker's Lodestone
	{ 87215, ST_Item, "Pandaria" },			-- Wormhole Generator: Pandaria
	
	{ 131225, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Setting Sun	
	{ 131222, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Mogu King
	{ 131231, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Scarlet Blade	
	{ 131229, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Scarlet Mitre	
	{ 131232, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Necromancer
	{ 131206, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Shado-Pan
	{ 131228, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Black Ox
	{ 131205, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Stout Brew
	{ 131204, ST_Challenge, "Pandaria Dungeons" },	-- Path of the Jade Serpent
	
	{ 147420, ST_Spell, "Random" },			-- One With Nature
	{ 64457, ST_Item, "Random" }, 			-- The Last Relic of Argus
		
	{ 33690, ST_Spell, "Shattrath" },		-- Teleport: Shattrath (Alliance)
	{ 33691, ST_Spell, "Shattrath" },		-- Portal: Shattrath (Alliance)
	{ 35715, ST_Spell, "Shattrath" },		-- Teleport: Shattrath (Horde)
	{ 35717, ST_Spell, "Shattrath" },		-- Portal: Shattrath (Horde)
	
	{ 128353, ST_Item, "Shipyard" },		-- Admiral's Compass
	
	{ 32272, ST_Spell, "Silvermoon" },		-- Teleport: Silvermoon
	{ 32267, ST_Spell, "Silvermoon" },		-- Portal: Silvermoon
	
	{ 49358, ST_Spell, "Stonard" },		-- Teleport: Stonard
	{ 49361, ST_Spell, "Stonard" },		-- Portal: Stonard
	
	{ 3561, ST_Spell, "Stormwind" },		-- Teleport: Stormwind
	{ 10059, ST_Spell, "Stormwind" },		-- Portal: Stormwind
	{ 63206, ST_Item, "Stormwind" },		-- Wrap of Unity
	{ 63352, ST_Item, "Stormwind" },		-- Shroud of Cooperation
	{ 65360, ST_Item, "Stormwind" },		-- Cloak of Coordination

	{ 128502, ST_Item, "Tanaan Jungle", MapIDTanaanJungle },	-- Hunter's Seeking Crystal
	{ 128503, ST_Item, "Tanaan Jungle", MapIDTanaanJungle },	-- Master Hunter's Seeking Crystal
	
	{ 18986, ST_Item, "Tanaris" },			-- Ultrasafe Transporter - Gadgetzan
	
	{ 49359, ST_Spell, "Theramore" },		-- Teleport: Theramore
	{ 49360, ST_Spell, "Theramore" },		-- Portal: Theramore
	
	{ 103678, ST_Item, "Timeless Isle" },	-- Time-Lost Artifact

	{ 3566, ST_Spell, "Thunder Bluff" },	-- Teleport: Thunder Bluff
	{ 11420, ST_Spell, "Thunder Bluff" },	-- Portal: Thunder Bluff
	
	{ 63378, ST_Item, "Tol Barad" },		-- Hellscream's Reach Tabard
	{ 63379, ST_Item, "Tol Barad" },		-- Baradin's Wardens Tabard
	{ 88342, ST_Spell, "Tol Barad" },		-- Teleport: Tol Barad (Alliance)
	{ 88344, ST_Spell, "Tol Barad" },		-- Teleport: Tol Barad (Horde)
	{ 88345, ST_Spell, "Tol Barad" },		-- Portal: Tol Barad (Alliance)
	{ 88346, ST_Spell, "Tol Barad" },		-- Portal: Tol Barad (Horde)
	

	{ 3563, ST_Spell, "Undercity" },		-- Teleport: Undercity
	{ 11418, ST_Spell, "Undercity" },		-- Portal: Undercity
	
	-- I don't know why there are so many of these, not sure which is right but it's now safe to
	-- list them all.
	{ 132621, ST_Spell, "Vale of Eternal Blossoms" },		-- Teleport: Vale of Eternal Blossoms
	{ 132627, ST_Spell, "Vale of Eternal Blossoms" },		-- Teleport: Vale of Eternal Blossoms
	{ 132620, ST_Spell, "Vale of Eternal Blossoms" },		-- Portal: Vale of Eternal Blossoms
	{ 132622, ST_Spell, "Vale of Eternal Blossoms" },		-- Portal: Vale of Eternal Blossoms
	{ 132624, ST_Spell, "Vale of Eternal Blossoms" },		-- Portal: Vale of Eternal Blossoms
	{ 132626, ST_Spell, "Vale of Eternal Blossoms" },		-- Portal: Vale of Eternal Blossoms
	
	{ 18984, ST_Item, "Winterspring" },	-- Dimensional Ripper - Everlook
}

local function GetOption(option)
	if TomeOfTele_Options == nil or TomeOfTele_Options[option] == nil then
		return DefaultOptions[option][1]
	else
	return TomeOfTele_Options[option]
	end
end

local function GetScale()
	return GetOption("scale") * UIParent:GetEffectiveScale()
end

local function GetScaledOption(option)
	return GetOption(option) * GetScale()
end

local function SetOption(option, value)
	if TomeOfTele_Options == nil then
		TomeOfTele_Options = {}
	end
	TomeOfTele_Options[option] = value
end

function Teleporter_OnEvent(self, event, ...)
	if event == "ADDON_LOADED" then
		local loadedAddon = ...
		if string.upper(loadedAddon) == string.upper(AddonName) then
			Teleporter_OnAddonLoaded()			
		end
	elseif event == "UNIT_SPELLCAST_SUCCEEDED" then
		local player, spell = ...
		if player == "player" then
			if spell == CastSpell then
				TeleporterClose()
			end
		end
	elseif event == "UNIT_INVENTORY_CHANGED" then
		TeleporterUpdateAllButtons()
	elseif event == "PLAYER_REGEN_DISABLED" then
		-- Can't close while in combat due to secure buttons, so disable Esc key
		if TeleporterParentFrame then
			local frameIndex = TeleporterFindInSpecialFrames()
			if frameIndex then
				tremove(UISpecialFrames,frameIndex);
			end
		end
	elseif event == "PLAYER_REGEN_ENABLED" then
		if TeleporterParentFrame then
			if not TeleporterFindInSpecialFrames() then
				tinsert(UISpecialFrames,TeleporterParentFrame:GetName());
			end
		end
	elseif event == "ZONE_CHANGED" or event == "ZONE_CHANGED_NEW_AREA" or event == "ZONE_CHANGED_INDOORS" then
		TeleporterCheckItemsWereEquiped()
	end
end

function TeleporterFindInSpecialFrames()
	for i,f in ipairs(UISpecialFrames) do
		if f == TeleporterParentFrame:GetName() then
			return i
		end
	end
	return nil
end

function Teleporter_OnLoad() 
	SlashCmdList["TELEPORTER"] = TeleporterSlashCmdFunction
	SLASH_TELEPORTER1 = "/tomeofteleport"
	SLASH_TELEPORTER2 = "/tele"	

	SlashCmdList["TELEPORTEREQUIP"] = TeleporterEquipSlashCmdFunction
	SLASH_TELEPORTEREQUIP1 = "/teleporterequip"

	SlashCmdList["TELEPORTERUSEITEM"] = TeleporterUseItemSlashCmdFunction
	SLASH_TELEPORTERUSEITEM1 = "/teleporteruseitem"

	SlashCmdList["TELEPORTERCASTSPELL"] = TeleporterCastSpellSlashCmdFunction
	SLASH_TELEPORTERCASTSPELL1 = "/teleportercastspell"

	SlashCmdList["TELEPORTERCREATEMACRO"] = TeleporterCreateMacroSlashCmdFunction
	SLASH_TELEPORTERCREATEMACRO1 = "/teleportercreatemacro"
end 

local function SavePosition()
	TomeOfTele_Points = {}
	for i = 1,TeleporterParentFrame:GetNumPoints(),1 do
		tinsert(TomeOfTele_Points,{TeleporterParentFrame:GetPoint(i)})
	end
end


local function Refresh()
	if IsVisible then 
		TeleporterClose()
		TeleporterOpenFrame()
	end
end

local TeleporterMenu = nil

local function OnHideItems(info)
	TomeOfTele_HideItems = not TomeOfTele_HideItems
	info.checked = TomeOfTele_HideItems
	
	Refresh()
end

local function OnHideSpells(info)
	TomeOfTele_HideSpells = not TomeOfTele_HideSpells
	info.checked = TomeOfTele_HideSpells
	
	Refresh()
end

local function OnHideChallenges(info)
	TomeOfTele_HideChallenge = not TomeOfTele_HideChallenge
	info.checked = TomeOfTele_HideChallenge
	
	Refresh()
end

local function OnSortbyType(info)
	TomeOfTele_SortByType = not TomeOfTele_SortByType
	info.checked = TomeOfTele_SortByType
	
	Refresh()
end

local function TomeOfTele_SetScale(scale)
	SetOption("scale", scale)
	if IsVisible then		
		TeleporterClose()
		TeleporterOpenFrame()
	end	
end

local function InitTeleporterMenu(frame, level, menuList)
	if level == 1 then 
		local info = UIDropDownMenu_CreateInfo()
		info.text = "Hide Items"
		info.value = 1
		info.func = function(info) OnHideItems(info) end
		info.owner = TeleporterMenu
		info.checked = TomeOfTele_HideItems
		UIDropDownMenu_AddButton(info, 1)		
		
		info.text = "Hide Class Spells"
		info.value = 2
		info.func = function(info) OnHideSpells(info) end
		info.owner = TeleporterMenu
		info.checked = TomeOfTele_HideSpells
		UIDropDownMenu_AddButton(info, 1)	
		
		info.text = "Hide Challenge Mode Spells"
		info.value = 3
		info.func = function(info) OnHideChallenges(info) end
		info.owner = TeleporterMenu
		info.checked = TomeOfTele_HideChallenge
		UIDropDownMenu_AddButton(info, 1)	
		
		info.text = "Sort by Type"
		info.value = 4
		info.func = function(info) OnSortbyType(info) end
		info.owner = TeleporterMenu
		info.checked = TomeOfTele_SortByType
		UIDropDownMenu_AddButton(info, 1)	
		
		info.text = "Scale"
		info.hasArrow = true
		info.menuList = "Scale"
		info.value = 5
		UIDropDownMenu_AddButton(info, 1)	
	elseif menuList == "Scale" then
		local scales = { 60, 70, 80, 90, 100, 110, 120, 130, 140, 150, 160, 170, 180, 190, 200 }
		for i,s in ipairs(scales) do
			local info = UIDropDownMenu_CreateInfo()
			info.text = s
			info.value = s / 10
			info.func = function(info) TomeOfTele_SetScale(s / 100) end
			info.owner = TeleporterMenu
			info.checked = function(info) return GetOption("scale") == s / 100 end
			UIDropDownMenu_AddButton(info, level)	
		end
	end
end

local function ShowMenu()
	if not TeleporterMenu then
		TeleporterMenu = CreateFrame("Frame", "TomeOFTeleMenu", UIParent, "UIDropDownMenuTemplate")
		UIDropDownMenu_Initialize(TeleporterMenu, InitTeleporterMenu, "MENU")
	end
	
	ToggleDropDownMenu(1, nil, TeleporterMenu, "cursor", 3, -3)
end

function TeleporterItemMustBeEquipped(item)
	if IsEquippableItem( item ) then
		return not IsEquippedItem ( item )
	else
		return false
	end
end

function TeleporterUpdateButton(button)

	if UnitAffectingCombat("player") then
		return
	end

	local settings = ButtonSettings[button]
	local isItem = settings[1]
	
	local item = settings[2]
	local cooldownbar = settings[3]
	local cooldownString = settings[4]
	local itemId = settings[5]
	local countString = settings[6]
	local onCooldown = false
	local buttonInset = GetScaledOption("buttonInset")

	if item then
		local cooldownStart, cooldownDuration
		if isItem then
			cooldownStart, cooldownDuration = GetItemCooldown(itemId)
		else
			cooldownStart, cooldownDuration = GetSpellCooldown(itemId)
		end

		if cooldownStart and cooldownStart > 0 then
			if GetTime() < cooldownStart then
				-- Long cooldowns seem to be reported incorrectly after a server reset.  Looks like the
				-- time is taken from a 32 bit unsigned int.
				cooldownStart = cooldownStart - 4294967.295
			end

			onCooldown = true
			local durationRemaining = cooldownDuration - ( GetTime() - cooldownStart )
			
			local parentWidth = button:GetWidth()
			local inset = 8
			cooldownbar:SetWidth( inset + ( parentWidth - inset ) * durationRemaining / cooldownDuration )
			
			if durationRemaining > 3600 then
				cooldownString:SetText(string.format("%.0fh", durationRemaining / 3600))
			elseif durationRemaining > 60 then
				cooldownString:SetText(string.format("%.0fm", durationRemaining / 60))
			else
				cooldownString:SetText(string.format("%.0fs", durationRemaining))
			end
			
		else			
			cooldownString:SetText("")
			cooldownbar:SetWidth( 1 )
		end		
		
		cooldownString:SetPoint("TOPLEFT",button,"TOPRIGHT",-cooldownString:GetStringWidth()*1.1-buttonInset-2,-buttonInset)
		cooldownString:SetPoint("BOTTOMRIGHT",button,"BOTTOMRIGHT",-buttonInset - 2,6)
		
		if countString and isItem then
			countString:SetText(GetItemCount(itemId))
		end

		if isItem and TeleporterItemMustBeEquipped( item ) then 
			button:SetBackdropColor(1,0,0,1)

			button:SetAttribute(
				"macrotext1",
				"/teleporterequip " .. item)
		elseif onCooldown then
			if cooldownDuration >2 then
				button:SetBackdropColor(1,0.7,0,1)
			else
				button:SetBackdropColor(0,0.7,0,1)
			end
			button:SetAttribute(
				"macrotext1",
				"/script print( \"" .. item .. " is currently on cooldown.\")")
		else
			button:SetBackdropColor(0,0.7,0,1)

			if isItem then
				button:SetAttribute(
					"macrotext1",
					"/teleporteruseitem " .. item .. "\n" ..
					"/use " .. item .. "\n" )
			else
				button:SetAttribute(
					"macrotext1",
					"/teleportercastspell " .. item .. "\n" ..
					"/cast " .. item .. "\n" )
			end
		end

		button:SetAttribute(
			"ctrl-macrotext1",
			"/teleportercreatemacro " .. item )
	end
end

function TeleporterUpdateAllButtons()	
	for button, settings in pairs(ButtonSettings) do
		TeleporterUpdateButton( button )
	end
end

function TeleporterShowItemTooltip( item, button )
	local _,link = GetItemInfo(item)
	if link then
		GameTooltip:SetOwner(button, "ANCHOR_BOTTOMRIGHT")
		GameTooltip:SetHyperlink(link)
		GameTooltip:Show()
	end
end

function TeleporterShowSpellTooltip( item, button )
	local link = GetSpellLink(item)
	if link then
		GameTooltip:SetOwner(button, "ANCHOR_BOTTOMRIGHT")
		GameTooltip:SetHyperlink(link)
		GameTooltip:Show()
	end
end

local function OnClickFrame(frame, button)
	if button == "RightButton" then
		ShowMenu()
	end
end

local function SortSpells(spell1, spell2)
	local spellId1 = spell1[1]
	local spellId2 = spell2[1]
	local spellName1 = spell1[SpellNameIndex]
	local spellName2 = spell2[SpellNameIndex]
	local spellType1 = spell1[2]
	local spellType2 = spell2[2]
	local zone1 = spell1[3]
	local zone2 = spell2[3]
	
	if TomeOfTele_SortByType then
		if spellType1 ~= spellType2 then
			return spellType1 < spellType2
		end
	end
	
	if zone1 ~= zone2 then
		return zone1 < zone2
	end
	
	return spellName1 < spellName2
end

local function SetupSpells()
	for index, spell in ipairs(TeleporterSpells) do		
		if spell[2] == ST_Item then
			spell[SpellNameIndex] = GetItemInfo( spell[1] )
		else
			spell[SpellNameIndex] = GetSpellInfo( spell[1] )			
		end
		
		if not spell[SpellNameIndex] then
			spell[SpellNameIndex] = "<Loading>"
		end
	end
end

function TeleporterOpenFrame()

	if UnitAffectingCombat("player") then
		print( "Cannot use " .. AddonTitle .. " while in combat." )
		return
	end
	
	if not IsVisible then			
		local buttonHeight = GetScaledOption("buttonHeight")
		local buttonWidth = GetScaledOption("buttonWidth")
		local labelHeight = GetScaledOption("labelHeight")
		local numColumns = 1
		local lastDest = nil
		local maximumHeight = GetScaledOption("maximumHeight")
		local fontHeight = GetScaledOption("fontHeight")
		local frameEdgeSize = 16 
		local fontFile = "Fonts\\FRIZQT__.TTF"
		local fontFlags = nil 
		local titleWidth = fontHeight * 28
		local titleHeight = fontHeight * 5
		local buttonInset = GetOption("buttonInset")

		IsVisible = true

		if TeleporterParentFrame == nil then
			TeleporterParentFrame = TeleporterFrame
			TeleporterParentFrame:SetFrameStrata("BACKGROUND")		
			TeleporterParentFrame:SetBackdrop({bgFile = "Interface/AchievementFrame/UI-Achievement-AchievementBackground", 
												edgeFile = "Interface/DialogFrame/UI-DialogBox-Border", 
												tile = false, edgeSize = frameEdgeSize, 
												insets = { left = buttonInset, right = buttonInset, top = buttonInset, bottom = buttonInset }});
			TeleporterParentFrame:ClearAllPoints()
			if TomeOfTele_Points then
				for i,pt in ipairs(TomeOfTele_Points) do
					TeleporterParentFrame:SetPoint(pt[1], pt[2], pt[3], pt[4], pt[5])
				end
			else
				TeleporterParentFrame:SetPoint("CENTER",0,0)
			end
			TeleporterParentFrame:SetBackdropColor(1,1,1,0.5)
			
			tinsert(UISpecialFrames,TeleporterParentFrame:GetName());
			--TeleporterParentFrame:SetScript( "OnHide", TeleporterClose )

			-- Title bar
			local titleFrame = CreateFrame("Frame","TeleporterTitleFrame",TeleporterParentFrame)	
			local titleFrameBG = titleFrame:CreateTexture()
			titleFrameBG:SetTexture("Interface/DialogFrame/UI-DialogBox-Header")
			titleFrameBG:SetAllPoints( titleFrame )
			titleFrame.texture = titleFrameBG
			titleFrame:SetPoint("TOP",TeleporterParentFrame,"TOP",0,titleHeight / 2 - 12 * GetScale())
			titleFrame:SetWidth(titleWidth)
			titleFrame:SetHeight(titleHeight)

			local titleString = titleFrame:CreateFontString("TeleporterTitleString", nil, "GameFontNormalSmall")
			titleString:SetFont(fontFile, fontHeight, fontFlags)
			titleString:SetText( AddonTitle )
			titleString:SetPoint("TOP", titleFrame, "TOP", 0, -titleHeight / 5)
			
			TeleporterParentFrame:RegisterForDrag("LeftButton")			
			TeleporterParentFrame:SetScript("OnDragStart", function() TeleporterParentFrame:StartMoving() end )
			TeleporterParentFrame:SetScript("OnDragStop", function() TeleporterParentFrame:StopMovingOrSizing(); SavePosition(); end )
			TeleporterParentFrame:EnableMouse(true)
			TeleporterParentFrame:SetMovable(true)
			TeleporterParentFrame:SetScript("OnMouseUp", OnClickFrame)
			
			-- Close button
			local closeButton = CreateFrame( "Button", "TeleporterCloseButton", TeleporterParentFrame, "UIPanelButtonTemplate" )
			closeButton:SetText( "X" )
			closeButton:SetPoint( "TOPRIGHT", TeleporterParentFrame, "TOPRIGHT", -buttonInset, -buttonInset )
			closeButton:SetWidth( buttonWidth )
			closeButton:SetHeight( buttonHeight )
			closeButton:SetScript( "OnClick", TeleporterClose )			
			
			-- Help text
			if GetOption("showHelp") then
				local helpString = TeleporterParentFrame:CreateFontString("TeleporterHelpString", nil, "GameFontNormalSmall")
				helpString:SetFont(fontFile, fontHeight, fontFlags)
				helpString:SetText( "Click to teleport, Ctrl+click to create a macro." )
				helpString:SetJustifyV("CENTER")
				helpString:SetJustifyH("LEFT")
			end
		end
		
		-- UI scale may have changed, resize
		TeleporterCloseButton:SetWidth( buttonHeight )
		TeleporterCloseButton:SetHeight( buttonHeight )		
		TeleporterCloseButtonText:SetFont(fontFile, fontHeight, fontFlags)
		if TeleporterHelpString then
			TeleporterHelpString:SetFont(fontFile, fontHeight, fontFlags)
		end
		TeleporterTitleFrame:SetWidth(titleWidth)
		TeleporterTitleFrame:SetHeight(titleHeight)
		TeleporterTitleFrame:SetPoint("TOP",TeleporterParentFrame,"TOP",0,titleHeight / 2 - 12 * GetScale())
		TeleporterTitleString:SetFont(fontFile, fontHeight, fontFlags)
		TeleporterTitleString:SetPoint("TOP", TeleporterTitleFrame, "TOP", 0, -titleHeight / 5)				

		local minyoffset = -buttonInset - 10
		local yoffset = minyoffset
		local maxyoffset = -yoffset
		local xoffset = buttonInset

		ButtonSettings = {}
		
		SetupSpells()
		table.sort(TeleporterSpells, SortSpells)

		for index, spell in ipairs(TeleporterSpells) do		
			local spellId = spell[1]
			local spellType = spell[2]
			local isItem = (spellType == ST_Item)
			local destination = spell[3]
			local zoneLimited = spell[4]
			local consumable = spell[5]
			local spellName = spell[SpellNameIndex]
			local isValidSpell = true
			local itemTexture = nil

			if destination == HearthString then
				local bindLocation = GetBindLocation()
				if bindLocation then
					destination = "Hearth (" .. bindLocation .. ")"
				else
					destination = "Hearth"
				end
			end
			
			-- Glyph of Astral Fixation changes spell destination to faction's Earthshrine
			local AstralGlyph = 147787
			if destination == RecallString then
				local hasGlyph = false
				local glyphIdx
				for glyphIdx = 1,NUM_GLYPH_SLOTS do
					local _,_,_,glyphId = GetGlyphSocketInfo(glyphIdx)
					if glyphId == AstralGlyph then
						hasGlyph = true
					end
				end
				
				if not hasGlyph then
					local bindLocation = GetBindLocation()
					if bindLocation then
						destination = "Hearth (" .. bindLocation .. ")"
					end
				else
					destination = "Astral Recall (Earthshrine)"
				end
			end

			if isItem then
				_, _, _, _, _, _, _, _, _, itemTexture = GetItemInfo( spellId )
				if not spellName then
					isValidSpell = false
				end
			else
				_,_,itemTexture = GetSpellInfo( spellId )
				if not spellName then
					isValidSpell = false
				end
			end
			
						
			local haveSpell = false
			if isValidSpell then
				if isItem then
					haveSpell = GetItemCount( spellId ) > 0
				else
					haveSpell = IsSpellKnown( spellId )
				end
			end
			
			if zoneLimited then
				local oldMapID = GetCurrentMapAreaID()
				SetMapToCurrentZone()
				local zoneID = GetCurrentMapAreaID()
				local continentID = GetCurrentMapContinent()
				SetMapByID(oldMapID)
							
				-- Continents are represented by negative values for zoneLimited
				if zoneLimited <= 0 then
					if -continentID ~= zoneLimited then
						haveSpell = false
					end
				else
					if zoneID ~= zoneLimited then
						haveSpell = false
					end
				end
			end
			
			-- Uncomment this to test all items.
			--haveSpell = true
			
			if TomeOfTele_HideItems and spellType == ST_Item then
				haveSpell = false
			end
			
			if TomeOfTele_HideSpells and spellType == ST_Spell then
				haveSpell = false
			end
			
			if TomeOfTele_HideChallenge and spellType == ST_Challenge then
				haveSpell = false
			end
			
			if haveSpell then
				-- Add extra column if needed
				local newColumn = false
				if -yoffset > maximumHeight then
					yoffset = minyoffset
					xoffset = xoffset + buttonWidth
					numColumns = numColumns + 1
					newColumn = true
				end		

				-- Title
				if newColumn or lastDest ~= destination then
					local destString = TeleporterCreateReusableFontString("TeleporterDL", TeleporterParentFrame, "GameFontNormalSmall")
					destString:SetFont(fontFile, fontHeight, fontFlags)
					destString:SetPoint("TOPLEFT", TeleporterParentFrame, "TOPLEFT", xoffset, yoffset)
					destString:SetPoint("BOTTOMRIGHT", TeleporterParentFrame, "TOPLEFT", buttonWidth + xoffset, yoffset - labelHeight)
					destString:SetText(destination)
					yoffset = yoffset - labelHeight
				end
				lastDest = destination	

				-- Main button
				local buttonFrame = TeleporterCreateReusableFrame("Button","TeleporterB",TeleporterParentFrame,"SecureActionButtonTemplate")
				buttonFrame:SetFrameStrata("BACKGROUND")
				buttonFrame:SetWidth(buttonWidth)
				buttonFrame:SetHeight(buttonHeight)
				buttonFrame:SetPoint("TOPLEFT",TeleporterParentFrame,"TOPLEFT",xoffset,yoffset)
				yoffset = yoffset - buttonHeight
				
				local buttonBorder = 4 * GetScale()
		
				buttonFrame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
													edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
													tile = true, tileSize = 16, edgeSize = 16 * GetScale(), 
													insets = { left = buttonBorder, right = buttonBorder, top = buttonBorder, bottom = buttonBorder }});
											
				buttonFrame:SetAttribute("type", "macro")
				buttonFrame:Show()

				if isItem then
					buttonFrame:SetScript(
						"OnEnter",
						function()
							TeleporterShowItemTooltip( spellName, buttonFrame )
						end )
				else
					buttonFrame:SetScript(
						"OnEnter",
						function()
							TeleporterShowSpellTooltip( spellName, buttonFrame )
						end )
				end

				buttonFrame:SetScript(
					"OnLeave",
					function()
						GameTooltip:Hide()
					end )

				-- Icon
				local teleicon = TeleporterCreateReusableFrame( "Frame", "TeleporterI", buttonFrame, nil )
				teleicon:SetPoint("TOPLEFT",buttonFrame,"TOPLEFT",6 * GetScale(),-5 * GetScale())
				if itemTexture then
					teleicon:SetWidth(buttonHeight - 10 * GetScale())
					teleicon:SetHeight(buttonHeight - 10 * GetScale())
					teleicon:SetBackdrop({bgFile = itemTexture})
				else
					teleicon:SetWidth(1)
					teleicon:SetHeight(1)
				end

				-- Cooldown bar
				local cooldownbar = TeleporterCreateReusableFrame( "Frame", "TeleporterCB", buttonFrame, nil )
				cooldownbar:SetFrameStrata("LOW")
				cooldownbar:SetWidth(64)
				cooldownbar:SetHeight(buttonHeight)
				cooldownbar:SetPoint("TOPLEFT",buttonFrame,"TOPLEFT",0,0)
				cooldownbar:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background",insets = { left = 4, right = 4, top = 3, bottom = 3 }})

				-- Cooldown label
				local cooldownString = TeleporterCreateReusableFontString("TeleporterCL", cooldownbar, "GameFontNormalSmall")
				cooldownString:SetFont(fontFile, fontHeight, fontFlags)
				cooldownString:SetJustifyH("RIGHT")
				cooldownString:SetJustifyV("CENTER")
				cooldownString:SetPoint("TOPLEFT",buttonFrame,"TOPRIGHT",-50,-buttonInset - 2)
				cooldownString:SetPoint("BOTTOMRIGHT",buttonFrame,"BOTTOMRIGHT",-buttonInset - 2,6)
				
				-- Name label
				local nameString = TeleporterCreateReusableFontString("TeleporterSNL", cooldownbar, "GameFontNormalSmall")
				nameString:SetFont(fontFile, fontHeight, fontFlags)
				nameString:SetJustifyH("LEFT")
				nameString:SetJustifyV("CENTER")
				nameString:SetPoint("TOP",cooldownString,"TOPRIGHT",0,0)
				nameString:SetPoint("LEFT",teleicon,"TOPRIGHT",2,-1)
				nameString:SetPoint("BOTTOMRIGHT",cooldownString,"BOTTOMLEFT",0,0)
				nameString:SetText( spellName )
				
				-- Count label
				local countString = nil
				if consumable then
					countString = TeleporterCreateReusableFontString("TeleporterCT", teleicon, "SystemFont_Outline_Small")
					countString:SetJustifyH("RIGHT")
					countString:SetJustifyV("CENTER")
					countString:SetPoint("TOPLEFT",teleicon,"TOPLEFT",0,0)
					countString:SetPoint("BOTTOMRIGHT",teleicon,"BOTTOMRIGHT",0,-2)
					countString:SetText("")
				end
			
				if -yoffset > maxyoffset then
					maxyoffset = -yoffset
				end
			
				ButtonSettings[buttonFrame] = { isItem, spellName, cooldownbar, cooldownString, spellId, countString }
			end	
		end
		
		local helpTextHeight		
		
		if TeleporterHelpString then
			if numColumns == 1 then
				helpTextHeight = 40
			else
				helpTextHeight = 10
			end
			TeleporterHelpString:SetPoint("TOPLEFT", TeleporterParentFrame, "TOPLEFT", 4 + buttonInset, -maxyoffset - 3 )
			TeleporterHelpString:SetPoint("RIGHT", TeleporterParentFrame, "RIGHT", -buttonInset, 0)
			TeleporterHelpString:SetHeight( helpTextHeight )	
		else
			helpTextHeight = 0
		end
		
		TeleporterParentFrame:SetWidth(numColumns * buttonWidth + buttonInset * 2)
		TeleporterParentFrame:SetHeight(maxyoffset + buttonInset * 2 + 2 + helpTextHeight)
		
	end

	TeleporterUpdateAllButtons()	
	TeleporterParentFrame:Show()
end


function TeleporterRestoreEquipment()
	ShouldNotBeEquiped = {}
	for slot,item in pairs(OldItems) do		
		ShouldNotBeEquiped[slot] = GetInventoryItemID("player", slot)
		ShouldBeEquiped[slot] = item
		EquipItemByName( item, slot )	
	end
	OldItems = {}
	EquipTime = GetTime()
end

function TeleporterCheckItemsWereEquiped()
	-- Sometimes equipping after casting fails. If that happens
	-- then try equipping after the next teleport.
	if GetTime() < EquipTime + 60 then 
		for slot, item in pairs(ShouldNotBeEquiped) do
			if IsEquippedItem ( item ) then
				EquipItemByName( ShouldBeEquiped[slot], slot )
			end
		end
	end
	ShouldNotBeEquiped = {}
	ShouldBeEquiped = {}
end

function TeleporterClose()
	if IsVisible and UnitAffectingCombat("player") then
		print( "Sorry, cannot close " .. AddonTitle .. " while in combat." )
	else
		if TeleporterParentFrame then
			TeleporterParentFrame:Hide()
			TeleporterHideCreatedUI()
			IsVisible = false
		end
		TeleporterRestoreEquipment()
	end
end

function TeleporterSlashCmdFunction(args)
	
	local splitArgs = {}
	if args then
		for v in string.gmatch(args, "[^ ]+") do
		  tinsert(splitArgs, v)
		end
	end
	
	if splitArgs[1] == "move" then
		local x = splitArgs[2]
		local y = splitArgs[3]
		if TeleporterParentFrame then
			TeleporterParentFrame:ClearAllPoints()
			TeleporterParentFrame:SetPoint("TOPLEFT", "UIParent", "TOPLEFT", x, -y)
			SavePoints()
		end
	elseif splitArgs[1] == "reset" then
		TomeOfTele_Points = nil
		if TeleporterParentFrame then
			TeleporterParentFrame:ClearAllPoints()
			TeleporterParentFrame:SetPoint("CENTER", "UIParent", "CENTER", 0, 0)
		end
	elseif splitArgs[1] == "showicon" then
		TomeOfTele_Icon.hide = false
		icon:Show("TomeTele")
	elseif splitArgs[1] == "hideicon" then
		TomeOfTele_Icon.hide = true
		icon:Hide("TomeTele")
	elseif splitArgs[1] == "set" then
		SetOption(splitArgs[2], splitArgs[3])
	elseif splitArgs[1] == nil then
		if IsVisible then
			TeleporterClose()
		else
			TeleporterOpenFrame()
		end	
	else
		print("Tome of Teleportation usage")
		print("/tele: Opens of closes the main window")
		print("/tele move x y: Move the window to a new location")
		print("/tele reset: Move the window back to the middle of the screen")
		print("/tele showicon: Show the minimap icon")
		print("/tele hideicon: Show the minimap icon")
		print("/tele set scale <scale>: Sets the scale of the window (default=1.0)")
	end
end

function TeleporterEquipSlashCmdFunction( item )
	CastSpell = nil

	if not IsEquippedItem ( item ) then
		if IsEquippableItem( item ) then 
			local _, _, _, _, _, _, _, _,itemEquipLoc = GetItemInfo(item)
			local itemSlot = InvTypeToSlot[ itemEquipLoc ]
			if itemSlot == nil then
				print( "Unrecognised equipable item type: " .. itemEquipLoc )
				return
			end
			OldItems[ itemSlot ] = GetInventoryItemID( "player", itemSlot )
			EquipItemByName( item, itemSlot )
		end
	end
end

function TeleporterUseItemSlashCmdFunction( item )
	local spell = GetItemSpell( item )
	TeleporterCastSpellSlashCmdFunction( spell )
end

function TeleporterCastSpellSlashCmdFunction( spell )
	CastSpell = spell
end

function TeleporterCreateMacroSlashCmdFunction( spell )
	if spell then
		local macro
		local printEquipInfo = false

		if GetItemInfo( spell ) then
			if IsEquippableItem( spell ) then
				macro =
					"#showtooltip " .. spell .. "\n" ..
					"/teleporterequip " .. spell .. "\n" ..
					"/teleporteruseitem " .. spell .. "\n" ..
					"/use " .. spell .. "\n"
				printEquipInfo = true
			else
				macro =
					"#showtooltip " .. spell .. "\n" ..
					"/use " .. spell .. "\n"
			end
		else
			macro =
				"#showtooltip " .. spell .. "\n" ..
				"/cast " .. spell .. "\n"
		end

		local macroName = "Use" .. string.gsub( spell, "[^%a%d]", "" )
		if GetMacroInfo( macroName ) then
			DeleteMacro( macroName )
		end
		CreateMacro( macroName, 1, macro, 1, 1 )

		local extraInstructions = ""
		if printEquipInfo then
			extraInstructions = "If the item is not equipped then the first click of the macro will equip it and the second click will use it."
		end
		print( "Created macro " .. macroName .. ". " .. extraInstructions )

		PickupMacro( macroName )
	end
end

function Teleporter_OnAddonLoaded()
	if TomeOfTele_Icon == nil then
		TomeOfTele_Icon = {}
	end
	
	icon:Register("TomeTele", dataobj, TomeOfTele_Icon)		
end

function Teleporter_OnUpdate()
	if IsVisible then
		TeleporterUpdateAllButtons()
	end
end

-----------------------------------------------------------------------
-- UI reuse

local uiElements = {}
local numUIElements = {}

-- Returns frame,frameName.  if frame is null then the caller must create a new object with this name
function TeleporterFindOrAddUIElement( prefix, parentFrame )
	local fullPrefix = parentFrame:GetName() .. prefix

	local numElementsWithPrefix = numUIElements[ fullPrefix ]
	if not numElementsWithPrefix then
		numElementsWithPrefix = 0
	end

	local frameName = fullPrefix .. numElementsWithPrefix
	local oldFrame = getglobal( frameName )
	if oldFrame then
		oldFrame:Show()
	end

	tinsert(uiElements, frameName)

	numElementsWithPrefix = numElementsWithPrefix + 1
	numUIElements[ fullPrefix ] = numElementsWithPrefix	
	
	return oldFrame, frameName
end


function TeleporterCreateReusableFrame( frameType, prefix, parentFrame, inheritsFrame )
	local frame, frameName = TeleporterFindOrAddUIElement( prefix, parentFrame )

	if not frame then
		frame = CreateFrame( frameType, frameName, parentFrame, inheritsFrame )
	end	
	
	return frame
end

function TeleporterCreateReusableFontString( prefix, parentFrame, font )
	local frame, frameName = TeleporterFindOrAddUIElement( prefix, parentFrame )

	if not frame then
		frame = parentFrame:CreateFontString(frameName, nil, font)
	end	
	
	return frame
end

function TeleporterHideCreatedUI()
	for index, itemName in pairs( uiElements ) do		
		local item = getglobal(itemName)
		if item then
			item:Hide()
		end
	end
	numUIElements = {}
	uiElements = {}
end

function dataobj:OnEnter()
    GameTooltip:SetOwner(self, "ANCHOR_NONE")
    GameTooltip:SetPoint("TOPLEFT", self, "BOTTOMLEFT")
    GameTooltip:ClearLines()

    GameTooltip:AddLine(AddonTitle, 1, 1, 1)

    GameTooltip:Show()
end

function dataobj:OnLeave()
    GameTooltip:Hide()
end

function dataobj:OnClick(button)
	if button == "LeftButton" then
		TeleporterSlashCmdFunction("")
	elseif button == "RightButton" then
		ShowMenu()
	end
end